--[[
	Todo

	Notes
		- hooking plates OnEnter requires the plates enable mouse. But doing that stops the mouseover ID being available. 
		I've tried disabing mouse once OnEnter fires but it's not guaranteed that the UPDATE_MOUSEOVER will then fire with the mobs moueoverID. This is more true when nameplates are clusterd ontop of each other.
		- TidyPlates sometimes removes the children from a plate.
		- TidyPlates seems to break my OnShow hook.
]]

local MAJOR, MINOR = "LibNameplate-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") return end

local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

lib.callbacks = lib.callbacks or LibStub("CallbackHandler-1.0"):New(lib)
if not lib.callbacks then	error(MAJOR .. " CallbackHandler-1.0.") return end

-- Globals
local _G					= _G
local WorldFrame_			= WorldFrame
local select_				= select
local UnitExists_			= UnitExists
local UnitGUID_				= UnitGUID
local UnitName_				= UnitName
local string_gsub			= string.gsub
local pairs_				= pairs
local table_remove			= table.remove
local table_insert			= table.insert
local GetMouseFocus_		= GetMouseFocus
local GetNumRaidMembers_	= GetNumRaidMembers
local GetNumPartyMembers_	= GetNumPartyMembers
local UnitIsUnit_			= UnitIsUnit
local GetRaidTargetIndex_	= GetRaidTargetIndex
local UnitInParty_			= UnitInParty
local tonumber_				= tonumber
local print_				= print
local math_floor			= math.floor
local RAID_CLASS_COLORS_	= RAID_CLASS_COLORS
local table_getn			= table.getn
local CreateFrame_			= CreateFrame

--Conts
local scanDelay		= 1 -- Scan for new nameplates every 1 seconds. We hook OnShow so this doesn't need to be done rapidly.
local updateDelay	= 1 -- Update nameplate info every 1 seconds. This checks if a nameplate has a raid icon then trys to find the GUID to that icon.


local DEBUG = false

--[===[@debug@
DEBUG = true
--@end-debug@]===]

local function CmdHandler()
	DEBUG = not DEBUG
end
_G.SlashCmdList["LIBNAMEPLATEDEBUG"] = CmdHandler
_G.SLASH_LIBNAMEPLATEDEBUG1 = "/lnbug"
local function debugPrint(...)
	if DEBUG == true then
		print_(...)
	end
end

local onShowHooks = {}
local onHideHooks = {}
local onUpdateHooks = {}

local regionNames = { "threatGlow", "healthBorder", "castBorder", "castNostop", "spellIcon", "highlightTexture", "nameText", "levelText", "dangerSkull", "raidIcon", "eliteIcon" }

--"Star", "Circle", "Diamond", "Triangle", "Moon", "Square", "Cross", "Skull"
--~ local raidIconTexture = "Interface\\TargetingFrame\\UI-RaidTargetingIcons"

--			= left							right
--			= ULx	ULy		LLx		LLy		URx		URy		LRx		LRy
-- star		= 0		0		0		0.25	0.25	0		0.25	0.25
-- circle	= 0.25	0		0.25	0.25	0.5		0		0.5		0.25
-- diamond	= 0.5	0		0.5		0.25	0.75	0		0.75	0.25
-- triangle	= 0.75	0		0.75	0.25	1		0		1		0.25
-- moon		= 0		0.25	0		0.5		0.25	0.25	0.25	0.5
-- square 	= 0.25	0.25	0.25	0.5		0.5		0.25	0.5		0.5
-- cross	= 0.5	0.25	0.5		0.5		0.75	0.25	0.75	0.5
-- skull	= 0.75	0.25	0.75	0.5		1		0.25	1		0.5
local raidIcons = {--from GetTexCoord. input is ULx and ULy (first 2 values).
	[0]		= {
		[0]		= 1, --star 
		[0.25]	= 5, --moon
	},
	[0.25]	= {
		[0]		= 2, --circle 
		[0.25]	= 6, --square
	},
	[0.5]	= {
		[0]		= 3, --star 
		[0.25]	= 7, --cross
	},
	[0.75]	= {
		[0]		= 4, --star 
		[0.25]	= 8, --skull
	},
}

lib.nameplates		= lib.nameplates or {}
lib.GUIDs			= lib.GUIDs or {}
lib.names			= lib.names or {}

local enableMouseFrames = {}

local plateGUIDs = {}

local colorToClass = {}
local function pctToInt(number) return math_floor((100*number) + 0.5) end
for classname, color in pairs_(RAID_CLASS_COLORS_) do
	colorToClass["C"..pctToInt(color.r)+pctToInt(color.g)+pctToInt(color.b)] = classname
--~ 	debugPrint("Class", classname, "C"..pctToInt(color.r)+pctToInt(color.g)+pctToInt(color.b))
end

local _ --yes a underscore, else it'll go global. =/

local function IsNamePlateFrame(frame)
	if frame:GetName() then
		debugPrint("IsNamePlateFrame","No name!")
		return false
	end

	if frame:GetID() ~= 0 then
		return false
	end
	
	if frame:GetObjectType() ~= "Frame" then
		debugPrint("IsNamePlateFrame","Not a frame!")
		return false
	end
		
	if frame:GetNumChildren() == 0 then
		debugPrint("IsNamePlateFrame","No children?!!", frame.extended and frame.extended.bars)
		return false
	end
		
	local children = {frame:GetChildren()}
	local sbFound = false
	for c = 1, table_getn(children) do
		if children[c]:GetObjectType() == "StatusBar" then
			sbFound = true
			break
		end
	end
	if sbFound == false then
		debugPrint("IsNamePlateFrame","No statusbar!")
		return false
	end

	return true
end

--[[
	these always return nil
	frame:IsClampedToScreen(), frame:IsIgnoringDepth(), frame:IsMovable(), frame:IsToplevel(), frame:GetDepth(), frame:GetEffectiveDepth(), 

]]

--[===[@debug@
function LC_TEST()--		/run LC_TEST()
	local c = {WorldFrame_:GetChildren()}
	local frame, cFrame
	for i = 1, table_getn(c) do
		frame = c[i]
		if frame:IsShown() then
			--frame.extended
			debugPrint(IsNamePlateFrame(frame), "ID:"..tostring(frame:GetID()), "C:"..tostring(frame:GetNumChildren()), "R:"..tostring(frame:GetNumRegions()))
			
		end
	end
end
--@end-debug@]===]

local isOnScreen = {}

local function FindNameplates(...)
--~ 	debugPrint("FindNameplates ")
	local frames = {WorldFrame_:GetChildren()}
	local frame
	for i = 1, table_getn(frames) do
		frame = frames[i]
		if frame:IsShown() then
			if IsNamePlateFrame(frame) then
				if not lib.nameplates[frame] then
					lib:SetupNameplate(frame, true)
				elseif not isOnScreen[frame] then
					debugPrint("The hell? Did our OnShow hook fail?")
					onShowHooks[frame] = false
					lib:SetupNameplate(frame, true)
				end
			end
		end
	end
--~ 	debugPrint("FindNameplates")
end

local function threatByColor( object , name)
	local redCan, greenCan, blueCan, alphaCan = object:GetVertexColor()
	local shown
	if not object:IsShown() then return "LOW" end
	if greenCan > .7 then return "MEDIUM" end
	if redCan > .7 then return "HIGH" end
end

local function reactionByColor(object)																											
	local redCan, greenCan, blueCan, alphaCan = object:GetStatusBarColor()
	if redCan < .01 and blueCan < .01 and greenCan > .99 then return "FRIENDLY", "NPC" 
	elseif redCan < .01 and blueCan > .99 and greenCan < .01 then return "FRIENDLY", "PLAYER"
	elseif redCan > .99 and blueCan < .01 and greenCan > .99 then return "NEUTRAL", "NPC"
	else return "HOSTILE", "PLAYER" end
end

local function combatByColor(r, g, b, a) 
	return (r > .5 and g < .5)
end

function lib:GetName(frame)
	return self.nameplates[frame] and string_gsub(self.nameplates[frame].regions.nameText:GetText(), "|cffffffff", "")
end

function lib:GetLevel(frame)
	return self.nameplates[frame] and self.nameplates[frame].regions.levelText:GetText()
end

function lib:GetReaction(frame)
	return self.nameplates[frame] and reactionByColor(self.nameplates[frame].bars.health)
end

function lib:GetType(frame)
	return self.nameplates[frame] and select_(2, reactionByColor(self.nameplates[frame].bars.health))
end

function lib:IsBoss(frame)
	return self.nameplates[frame] and self.nameplates[frame].regions.dangerSkull:IsShown()
end

function lib:GetHealthMax(frame)
	return self.nameplates[frame] and self.nameplates[frame].bars.health:GetMinMaxValues()
end

function lib:GetClass(frame)
--This will return nil if we're not in a PvP zone (like in citys)
	if self.nameplates[frame] then
		local red, green, blue = self.nameplates[frame].bars.health:GetStatusBarColor()
--~ 		debugPrint("GetClass", red, green, blue, "=", "C"..pctToInt(red)+pctToInt(green)+pctToInt(blue))
		return colorToClass["C"..pctToInt(red)+pctToInt(green)+pctToInt(blue)]
	end
end

function lib:IsElite(frame)
	return self.nameplates[frame] and self.nameplates[frame].regions.eliteIcon:IsShown()
end

function lib:GetThreatSituation(frame)
	return self.nameplates[frame] and threatByColor(self.nameplates[frame].regions.threatGlow)
end

function lib:GetAlpha(frame)-- really?
	return frame:GetAlpha()
end

function lib:IsTarget(frame)
	return (frame:GetAlpha() == 1 and UnitExists_("target"))
end

function lib:GetHealth(frame)
	return self.nameplates[frame] and self.nameplates[frame].bars.health:GetValue()
end

function lib:IsMouseover(frame)
	return self.nameplates[frame] and self.nameplates[frame].regions.highlightTexture:IsShown()
end

function lib:IsCasting(frame)
	return self.nameplates[frame] and self.nameplates[frame].bars.cast:IsShown()
end

function lib:IsInCombat(frame)
	return self.nameplates[frame] and combatByColor(self.nameplates[frame].regions.nameText:GetTextColor())
end

function lib:IsMarked(frame)
	return self.nameplates[frame] and self.nameplates[frame].regions.raidIcon:IsShown()
end

function lib:GetRaidIcon(frame)
	if self.nameplates[frame] then
		if self.nameplates[frame].regions.raidIcon:IsShown() then
			local ULx,ULy = self.nameplates[frame].regions.raidIcon:GetTexCoord() --,LLx,LLy,URx,URy,LRx,LRy
			if ULx and ULy then
				return raidIcons[ULx] and raidIcons[ULx][ULy]
			end
		end
	end
	return 0
end

function lib:GetGUID(frame)
	return plateGUIDs[frame]
end

-- ***
function lib:GetTargetNameplate()
	if self.targeted and self.nameplates[self.targeted] and self.nameplates[self.targeted].active == true then
		return self.targeted
	end
end

function lib:GetNameplateByGUID(GUID)
	if self.GUIDs[GUID] and self.nameplates[self.GUIDs[GUID]] and self.nameplates[self.GUIDs[GUID]].active == true then
		return self.GUIDs[GUID]
	end
end


function lib:GetNameplateByName(name)
	if self.names[name] and self.nameplates[self.names[name]] and self.nameplates[self.names[name]].active == true then
		return self.names[name]
	end
end

function lib:GetNameplateByUnit(unitID)
	if UnitIsUnit_(unitID, "target") then
		return self:GetTargetNameplate()
	end
	local GUID = UnitGUID_(unitID)
	if self.GUIDs[GUID] then
		return self.GUIDs[GUID], self.GUIDs[GUID].LibNP
	end
	local name = UnitName_(unitID)
	if self.names[GUID] then
		return self.names[GUID], self.names[GUID].LibNP
	end
end

local function RecycleNameplate(frame)
	lib.callbacks:Fire("LibNameplate_RecycleNameplate", frame)
	local LibNP = lib.nameplates[frame]
	if LibNP then
		local GUID = plateGUIDs[frame]
		local name = LibNP.unit.name
		
		if lib.GUIDs[GUID] then lib.GUIDs[GUID] = nil end
		if lib.names[name] then lib.names[name] = nil end
	end

	plateGUIDs[frame] = false
	lib.nameplates[frame].active = false
	
--~ 	debugPrint("RecycleNameplate", onShowHooks[frame], onShowHooks[frame]:GetParent(), onShowHooks[frame]:GetParent() == frame)
--~ 	onShowHooks[frame]:Show()
end

local function GetMouseoverGUID(frame)
	local unitID = "mouseover"
	if UnitExists_(unitID) then
		local GUID = UnitGUID_(unitID)
		plateGUIDs[frame] = GUID
		lib.GUIDs[GUID] = frame

		lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, unitID)
	end
end

local function FindPlateWithRaidIcon(iconNum, unitID)
	local GUID = UnitGUID_(unitID)
	local LibNP
	for frame, LibNP in pairs_(lib.nameplates) do 
		if lib:IsMarked(frame) and lib:GetRaidIcon(frame) == iconNum then
			return frame, LibNP
		end
	end
	return nil
end

local function FindGUIDByRaidIcon(frame, raidNum)
	local LibNP = lib.nameplates[frame]
	local group, num = "", 0
	if GetNumRaidMembers_() > 1 then
		group, num = "raid", GetNumRaidMembers_()
	elseif GetNumPartyMembers_() > 0 then
		group, num = "party", GetNumPartyMembers_()
	end
	
	local unitID, targetIndex, GUID
	for i = 1, num do
		unitID = "party"..i;
		if not UnitIsUnit_("target", unitID.."target") then
			targetIndex = GetRaidTargetIndex_(unitID.."target")
			if targetIndex and targetIndex == raidNum then
				GUID = UnitGUID_(unitID.."target")
				plateGUIDs[frame] = GUID
				
				lib.GUIDs[GUID] = frame
				lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, unitID.."target")
				return
			end
		end
	end
end

local function UpdateNameplateInfo(frame)
	local LibNP = lib.nameplates[frame]
	local unit = LibNP.unit
	local regions = LibNP.regions

--~ 	debugPrint("UpdateNameplateInfo", lib:GetName(frame))
	
	if not plateGUIDs[frame] and regions.highlightTexture:IsShown() then
		GetMouseoverGUID(frame)
	end
	
	--So I don't forget, we keep checking the raid icon so try to find the GUID. 
	if not plateGUIDs[frame] then
		local raidNum = lib:GetRaidIcon(frame)
		if raidNum and raidNum > 0 then
			FindGUIDByRaidIcon(frame, raidNum)
		end
	end
	
--~ 	debugPrint("Updated nameplate ",unit.name)
	frame.lnpLastUpdate = 0
end

local function ourOnShow(self, ...)
	lib:SetupNameplate(self)
end

local function ourOnHide(self, ...)
	isOnScreen[self] = false
	RecycleNameplate(self)
end

local function ourOnUpdate(self, elapsed, ...)
	if self.lnpCheckForTarget == true then --Check on the first OnUpdate after the frame's shown.
		self.lnpCheckForTarget = false
		if not plateGUIDs[self] and self:IsShown() and ((self:GetAlpha() == 1) and UnitExists_("target")) then
			

			local GUID = UnitGUID_("target")
			plateGUIDs[self] = GUID
			
			
			lib.GUIDs[GUID] = self
			lib.callbacks:Fire("LibNameplate_FoundGUID", self, GUID, "target")
		end
	end
	
	self.lnpLastUpdate = self.lnpLastUpdate + elapsed
	if self.lnpLastUpdate > updateDelay then
		UpdateNameplateInfo(self, true)
	end
end

--[[
local onEnterHooks = {}
local onLeaveHooks = {}
local lastMouseoverFrame = false --current frame we're moused over.
local function ourOnEnter(self, ...)
	if lastMouseoverFrame and lastMouseoverFrame ~= self then
		enableMouseFrames[lastMouseoverFrame] = true
		lib.frame.enableMouse = true
	end
	self:EnableMouse(false)--Now that our OnEnter has fired, disable mouse so that UPDATE_MOUSEOVER_UNIT fires.
end

Note to self, OnLeave will fire before UPDATE_MOUSEOVER_UNIT fires. 
local function ourOnLeave(self, ...)
	lastMouseoverFrame = self
	debugPrint("ourOnLeave", lastMouseoverFrame)
end]]


function lib:SetupNameplate(frame)
	isOnScreen[frame] = true
	if not lib.nameplates[frame] then
		lib.nameplates[frame] = {}
		local LibNP = lib.nameplates[frame]
		LibNP.frame = frame
		
		--REGIONS
		if not LibNP.regions then
			local regs = {frame:GetRegions()}
			LibNP.regions = {}
			
			for i, name in pairs_(regionNames) do 
				LibNP.regions[name] = regs[i]
			end
		end
		local regions = LibNP.regions
		
		--BARS
		LibNP.bars = LibNP.bars or {}
		local bars = LibNP.bars
		bars.health, bars.cast = frame:GetChildren()
		
		--UNIT
		LibNP.unit = LibNP.unit or {}


		
		frame.lnpLastUpdate = 0
	end

	if frame:HasScript("OnHide") and not onHideHooks[frame] then
		onHideHooks[frame] = true
		frame:HookScript("OnHide", ourOnHide)
	end
	
	if frame:HasScript("OnShow") and not onShowHooks[frame] then
		onShowHooks[frame] = true
		frame:HookScript("OnShow", ourOnShow)
	end
	
	if frame:HasScript("OnUpdate") and not onUpdateHooks[frame] then
		onUpdateHooks[frame] = true
		frame:HookScript("OnUpdate", ourOnUpdate)
	end
	
	
	--[[
	if frame:HasScript("OnEnter") and not onEnterHooks[frame] then
		onEnterHooks[frame] = true
		frame:HookScript("OnEnter", ourOnEnter)
		frame:EnableMouse(true) --Enable mouse to allow our OnEnter handler to fire.
	end

	if frame:HasScript("OnLeave") and not onLeaveHooks[frame] then
		onLeaveHooks[frame] = true
		frame:HookScript("OnLeave", ourOnLeave)
	end
	]]
	
	local LibNP = lib.nameplates[frame]
	local regions = LibNP.regions

	if frame:GetNumChildren() == 0 then
		debugPrint("SetupNameplate 1", "Where'd all the children go?", frame.extended, frame.extended.bars)
	end
	
	if frame:GetNumRegions() == 0 then
--~ 		debugPrint("SetupNameplate 1", "Where'd all the regions go?", frame.extended, frame.extended.regions)
		if _G.TidyPlates and not regions.tinyPlateRegions then

			if frame.extended and frame.extended.regions then
				
				for a, b in pairs_(frame.extended.regions) do 
	--~ 				debugPrint("SetupNameplate 2", a, b)
					
					regions[a] = b
				end
				regions.tinyPlateRegions = true
			end
--~ 		elseif _G.Aloft and not regions.aloftRegions then
		else
			return
		end
	else
		--confirm we still have the regions
		for regName, obj in pairs_(regions) do 
			if not obj.GetWidth then
				debugPrint("lost a reg?", regName)
			end
		end
		
	end

	local plateName = self:GetName(frame)

	local unit = LibNP.unit
	unit.name = plateName
	lib.names[plateName] = frame
	
	LibNP.active = true
	lib.callbacks:Fire("LibNameplate_NewNameplate", frame)
--~ 	--~ 		debugPrint("LibNameplate_NewNameplate", LibNP:GetName())
	frame.lnpCheckForTarget = true
	UpdateNameplateInfo(frame)
end--	/lnbug




local function CheckForTargetGUID()
	local unitID = "target"
	local GUID
	local highlightTexture, nameTextRegion
	for frame, LibNP in pairs_(lib.nameplates) do 
	
		if frame:IsShown() and frame:GetAlpha() == 1 and UnitExists_(unitID) then
			lib.targeted = frame
			if not plateGUIDs[frame] then
				GUID = UnitGUID_(unitID)
				plateGUIDs[frame] = GUID
				lib.GUIDs[GUID] = frame
				lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, unitID)
				
--~ 				_, _, _, _, _, highlightTexture, nameTextRegion = frame:GetRegions()
--~ 				debugPrint("Found target GUID", UnitName_(unitID), GUID, UnitName_(unitID) == nameTextRegion:GetText())
			end
			lib.callbacks:Fire("LibNameplate_TargetNameplate", frame)
			return
		end
	end
--~ 	debugPrint("Failed to find target GUID")
end



local function OnUpdate(self, elapsed)
	if self.checkTarget == true then
		self.checkTarget = false
		CheckForTargetGUID()
	end

	self.lastUpdate = self.lastUpdate + elapsed
	if self.lastUpdate >= scanDelay then
		self.lastUpdate = 0
		FindNameplates(WorldFrame_:GetChildren())
	end
end



local function OnEvent(self, event, ...)
	if event == "UPDATE_MOUSEOVER_UNIT" then
		
		--[[if GetMouseFocus_():GetName() == "WorldFrame" then
			if lastMouseoverFrame then
				local LibNP = lib.nameplates[lastMouseoverFrame]
				if not LibNP.unit.GUID then
					local GUID = UnitGUID_("mouseover")
					LibNP.unit.GUID = GUID
					lib.GUIDs[GUID] = lastMouseoverFrame
					lib.callbacks:Fire("LibNameplate_FoundGUID", lastMouseoverFrame, GUID, targetID)
					
				end
				enableMouseFrames[lastMouseoverFrame] = true
				self.enableMouse = true
				lastMouseoverFrame = false
			end
		else
			debugPrint(event, GetMouseFocus_():GetName())
		end]]

		if GetMouseFocus_():GetName() == "WorldFrame" then
			for frame, LibNP in pairs_(lib.nameplates) do 
				
				if frame:IsShown() and LibNP.regions.highlightTexture:IsShown() then
					if not plateGUIDs[frame] then
						GetMouseoverGUID(frame)
					end
					lib.callbacks:Fire("LibNameplate_MouseoverNameplate", frame)
					return
				end
			end
		end
		
	elseif event == "PLAYER_TARGET_CHANGED" then
		if UnitExists_("target") then
			self.checkTarget = true--Check for target nameplate on next OnUpdate. We can't do it now.
		else
			lib.targeted = nil
		end
		
	elseif event == "UNIT_TARGET" then
		local unitID = ...
		local targetID = unitID.."target"
		if not UnitIsUnit_("player", unitID) then
			local iconNum = GetRaidTargetIndex_(targetID)
			if iconNum and iconNum > 0 then
				local foundPlate = FindPlateWithRaidIcon(iconNum, targetID)
				local LibNP = lib.nameplates[foundPlate]
				if foundPlate and not plateGUIDs[foundPlate] then
					local GUID = UnitGUID_(targetID)
					plateGUIDs[foundPlate] = GUID
					lib.GUIDs[GUID] = foundPlate
					lib.callbacks:Fire("LibNameplate_FoundGUID", foundPlate, GUID, targetID)
				end
			end
		end
		
	elseif event == "RAID_TARGET_UPDATE" then
		local regions
		for frame, LibNP in pairs_(lib.nameplates) do 
			if frame:IsShown() and not plateGUIDs[frame] and lib:IsMarked(frame) then
				local raidNum = lib:GetRaidIcon(frame)
				if raidNum and raidNum > 0 then
					FindGUIDByRaidIcon(frame, raidNum)
				end
			end
		end
	end
end


if not lib.frame then
	lib.frame=CreateFrame_("Frame")
end

lib.frame.lastUpdate = 0
lib.frame.checkTarget = false
lib.frame:SetScript("OnUpdate", OnUpdate)
lib.frame:SetScript("OnEvent", OnEvent)
lib.frame:RegisterEvent('UPDATE_MOUSEOVER_UNIT')
lib.frame:RegisterEvent('PLAYER_TARGET_CHANGED')
lib.frame:RegisterEvent('UNIT_TARGET')
lib.frame:RegisterEvent('RAID_TARGET_UPDATE')